"""
Downsampling of 3D featuremaps in form T[batch_size, Z, feature, X, Y] by
subsample=(kx, ky, kz) to T2[batch_size, Z/kx, feature, X/ky, Y/kz]
Josef Pihera, 2012
"""
import theano
import theano.tensor as tensor
import numpy
import theano.tensor.signal.downsample as downsample

def sumZ(T):
    s = T.shape[1]
    i1 = tensor.arange(0, s, 2)
    i2 = tensor.arange(1, s, 2)
    Tf = T.dimshuffle(1,0,2,3,4)
    Tsum = Tf[i1] + Tf[i2]
    return Tsum.dimshuffle(1,0,2,3,4)

def sumX(T):
    s = T.shape[3]
    i1 = tensor.arange(0, s, 2)
    i2 = tensor.arange(1, s, 2)
    Tf = T.dimshuffle(3,0,1,2,4)
    Tsum = Tf[i1] + Tf[i2]
    return Tsum.dimshuffle(1,2,3,0,4)

def sumY(T):
    s = T.shape[4]
    i1 = tensor.arange(0, s, 2)
    i2 = tensor.arange(1, s, 2)
    Tf = T.dimshuffle(4,0,1,2,3)
    Tsum = Tf[i1] + Tf[i2]
    return Tsum.dimshuffle(1,2,3,4,0)


def downsample3d(T, subsample=(2,2,2), ignore_border_3d = True):
    #avg version
    return sumY(sumX(sumZ(T))) / 8.0
    
"""
    #max version 2
    #this works with max downsampling only
    #T[n, Z, f, X, Y]
    TnfZXY = T.dimshuffle(0,2,1,3,4)
    #TnfZXY[n , f, Z, X, Y]
    Tr = T.reshape((T.shape[0] * T.shape[2], T.shape[1], T.shape[3], T.shape[4]))
    #Tr[n * f, Z, X, Y]
    Td1 = downsample.max_pool_2d(input=Tr, ds=(subsample[0],subsample[1]),\
                                 ignore_border = ignore_border_3d)
    #Td1[n * f, Z, X/kx, Y/ky]
    Tr2 = Td1.dimshuffle(0,2,3,1)
    #Tr2[n *f, X/kx, Y/ky, Z]                                 
    Td0 = Tr2#downsample.max_pool_2d(input=Tr2, ds=(1, subsample[2]),\
          #                        ignore_border = ignore_border_3d)
    #Td0[n*f,X/kx,Y/ky,Z/kz]
    Tr3 = Td0.reshape((T.shape[0], T.shape[2], Td0.shape[1], Td0.shape[2], Td0.shape[3]))
    #Tr3[n, f, X/kx, Y/ky, Z/kz]
    return Tr3.dimshuffle(0,4,1,2,3)
"""
"""
    #max version 1
    #T[n, Z, f, X, Y]
    Tr = T.reshape((T.shape[0], T.shape[1] * T.shape[2], T.shape[3], T.shape[4]))
    #Tr[n, Z*f, X, Y]
    Td1 = downsample.max_pool_2d(input=Tr, ds=(subsample[0],subsample[1]),\
                                 ignore_border = ignore_border_3d)
    #Td1[n, Z * f, X/kx, Y/ky]                                 
    Td1r = Td1.reshape((T.shape[0], T.shape[1], T.shape[2], Td1.shape[2], Td1.shape[3]))
    #Td1r[n, Z, f, X/kx, Y/ky]
    Td1s = Td1r.dimshuffle(0,2,3,4,1)
    #Td1s[n,f,X/kx,Y/ky,Z]
    Td1sr = Td1s.reshape((T.shape[0], T.shape[2] * Td1.shape[2], Td1.shape[3], T.shape[1]))
    #Td1sr[n,f*X/kx,Y/ky,Z]
    Td0s = downsample.max_pool_2d(input=Td1sr, ds=(1,subsample[2]),\
                                  ignore_border = ignore_border_3d)
    #Td0s[n,f*X/kx,Y/ky,Z/kx]
    Td0r = Td0s.reshape((T.shape[0], T.shape[2], Td1.shape[2], Td1.shape[3], Td0s.shape[3]))
    #Td0r[n,f,X/kx,Y/ky,Z/kx]
    Td0 = Td0r.dimshuffle(0,4,1,2,3)
    return Td0
"""

def test():
    T = theano.shared(numpy.arange(64).reshape((1,4,1,4,4)))
    T2 = theano.shared(numpy.arange(8).reshape((1,2,1,2,2)))
    #current avg version should return 10.5, 12.5, 18.5, 20.5, 42.5, 44.5, 50.5, 52.5
    #and the T2 - 3.5
    #max version should return 21, 23, 29, 31, ,53, 55, 61, 63
    return downsample3d(T), downsample3d(T2)